from datetime import datetime, timezone

class AverageTime:
    DEFAULT_ID = "default"

    def __init__(self) -> None:
        self.data:dict[str, dict[str, list[datetime]]] = {}

    def __add_time__(self, start_time:datetime=None, end_time:datetime=None, id_time:str=DEFAULT_ID) -> None:
        if id_time not in list(self.data.keys()):
            self.data[id_time] = { "st": [], "ed": [] }

        if start_time is not None:
            self.data[id_time]["st"].append(start_time)

        if end_time is not None:
            self.data[id_time]["ed"].append(end_time)

    def __get_average_by_id__(self, id_time:str) -> None | str:
        assert id_time in list(self.data.keys()), "ID not found"
        if len(self.data[id_time]["st"]) == 0:
            return "E: Start time not found for id: " + id_time
        if len(self.data[id_time]["ed"]) == 0:
            return "E: Not enough value recorded for id: " + id_time
        if len(self.data[id_time]["st"])-1 > len(self.data[id_time]["ed"]):
            return "E: Too many stopwatches started"
        average_list:list[float] = []
        for index in range(len(self.data[id_time]["ed"])):
            stts:float = self.data[id_time]["st"][index].timestamp()
            edts:float = self.data[id_time]["ed"][index].timestamp()
            average_list.append(edts - stts)
        average:float = sum(average_list) / len(average_list)
        average:datetime = datetime.fromtimestamp(average)
        average -= datetime(year=1970, month=1, day=1, hour=1)
        return average

    def start(self, id:str=DEFAULT_ID) -> None:
        self.__add_time__(start_time=datetime.now(timezone.utc), 
                          id_time=id)

    def loop(self, id:str=DEFAULT_ID) -> None:
        t = datetime.now(timezone.utc)
        if id not in list(self.data.keys()):
            self.__add_time__(start_time=t, id_time=id)
        else:
            self.__add_time__(start_time=t, end_time=t, id_time=id)

    def stop(self, id:str=DEFAULT_ID) -> None:
        self.__add_time__(end_time=datetime.now(timezone.utc), 
                          id_time=id)

    def get_average(self, id:str=None) -> datetime | dict[str, datetime] | str:
        if id is None and [self.DEFAULT_ID] == list(self.data.keys()):
            return self.__get_average_by_id__(self.DEFAULT_ID)
        elif id is None:
            d = {key_id:self.__get_average_by_id__(key_id) for key_id in list(self.data.keys())}
            if len(list(d.keys())) > 1:
                return d
            else:
                return list(d.values())[0]
        else:
            return self.__get_average_by_id__(id)

    def __str__(self) -> str:
        avg = self.get_average()
        if isinstance(avg, dict):
            max_key_length = max([len(k) for k in list(avg.keys())])
            timelist = [k+(" "*(max_key_length-len(k)))+": "+str(v) for k, v in avg.items()]
            return "\n".join(timelist)
        else:
            return str(avg)

    def __repr__(self) -> str:
        return self.__str__()



CHARACTER_MANGA = """Sakurajima Mai
Azusagawa Sakuta
Koga Tomoe
Futaba Rio
Azusagawa Kaede
Toyohama Nodoka
Kunimi Yûma
Kamisato Saki
Makinohara Shôko
Kano Kotomi
Oga Tatsumi
Kaiser de Emperana Beelzebub IV
Kunieda Aoi
Hildegarde
Alaindelon Bathin de Emuna
Furuichi Takayuki
Ômori Nene
Tôjô Hidetora
Kanzaki Hajime
Natsume Shintarô
Tanimura Chiaki
Hanazawa Yuka
Himekawa Tatsuya
En Ô
Yolda
Nanami Shizuka
Shiroyama Takeshi
Tenma Umatarô
Ochanomizu Hiroshi
A106
Ochanomizu Ran
Tsutsumi Motoko
Ban Shunsaku
Komi Shôko
Tadano Hitohito
Osana Najimi
Agari Himiko
Yadano Makeru
Yamai Ren
Ishigami Senkû
Ôki Taiju
Ogawa Yuzuriha
Shishiô Tsukasa
Kohaku
Chrome
Asagiri Gen
Suika
Kinrô
Ginrô
Ruri
Kaseki
Ishigami Byakuya
Lillian Weinberg
Shamil Volkov
Connie Lee
Darya Nikitina
Yakov Nikitin
Magma
Hyôga
Momiji Homura
Hokutozai Minami
Soyuz
Kokuyô
Loid Forger
Yor Forger
Anya Forger
Franky Franklin
Henry Henderson
Damian Desmond
Becky Blackbell
Sylvia Sherwood
Donovan Desmond
Yuri Briar
Fukuya no Onna Shujin
Camilla
Dominic
Millie
Sharon
Bond Forger
Emile Elman
Ewen Egeberg
Akabane Karma
Koro-sensei
Shiota Nagisa
Kayano Kaede
Karasuma Tadaomi
Irina Jelavic
Nakamura Rio
Terasaka Ryôma
Sugino Tomohito
Isogai Yûma
Maehara Hiroto
Okuda Manami
Kanzaki Yukiko
Asano Gakuhô
Okajima Taiga
Kurahashi Hinano
Ritsu
Hayami Rinka
Kataoka Meg
Red Eye
Horibe Itona
Shiro
Lovro
Okano Hinata
Mimura Kôki
Sugaya Sôsuke
Chiba Ryûnosuke
Takebayashi Kôtarô
Fuwa Yuzuki
Yada Tôka
Asano Gakushû
Yukimura Aguri
Hara Sumire
Hazama Kirara
Araki Teppei
Sakakibara Ren
Kimura Masayoshi
Hiro
Zero Two
Ichigo
Gorô
Zorome
Miku
Ikuno
Mitsuru
Futoshi
Kokoro
Hoshino Ai
Hoshino Aquamarine
Hoshino Ruby
Arima Kana
Kurokawa Akane
MEM-cho
Amemiya Gorô
Tendôji Sarina
Saitô Ichigo
Saitô Miyako
Gotanda Taishi
Kaburagi Masaya
Narushima Melt
Shiranui Frill
Kotobuki Minami
Sumi Yuki
Kumano Nobuyuki
Yuzaki Nasa
Yuzaki Tsukasa
Arisugawa Kaname
Arisugawa Aya
Kaginoji Chitose
Charlotte
Aurora
Edward Elric
Alphonse Elric
Rose
Cornello
Trisha Elric
Winry Rockbell
Van Hohenheim
Roy Mustang
Maes Hughes
Riza Hawkeye
Gracia Hughes
Shô Tucker
Scar
Lyra
Psiren
Fletcher Tringham
Russel Tringham
Lust
Gluttony
Heimans Breda
Jean Havoc
Vato Falman
Kain Fuery
King Bradley
Basque Grand
Envy
Alex Louis Armstrong
Pinako Rockbell
Maria Ross
Sheska
Solf J. Kimblee
Elicia Hughes
Sloth
Paninya
Izumi Curtis
Wrath
Frank Archer
Dante
Greed
Martel
Lujon
Barry the Chopper
Barry the Chopper (humain)
Sig Curtis
Prisoner 48
Nina Tucker
Alexander
Tim Marcoh
Denny Brosh
Dominic LeCoulte
Yoki
Grumman
Dolcetto
Roa
Selim Bradley
Mrs Bradley"""

CHARACTER_MANGA_LIST = CHARACTER_MANGA.split("\n")



import wmi
import win32com.client

#Function to get processor details
def get_processor_details():
    processor_details = []
    str_computer = "."
    obj_wmi_service = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    obj_swbem_services = obj_wmi_service.ConnectServer(str_computer, "root\cimv2")
    col_items = obj_swbem_services.ExecQuery("SELECT * FROM Win32_Processor")
    for obj_item in col_items:
        processor = {
            'name': obj_item.Name,
            'description': obj_item.description,
            'DeviceID': obj_item.DeviceID,
            'manufacturer': obj_item.Manufacturer,
            'processorID': obj_item.processorID,
            'systemName': obj_item.SystemName
        }
        processor_details.append(processor)
    return processor_details

#Function to get motherboard details
def get_motherboard_details():
    motherboard_details = []
    str_computer = "."
    obj_wmi_service = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    obj_swbem_services = obj_wmi_service.ConnectServer(str_computer, "root\cimv2")
    col_items = obj_swbem_services.ExecQuery("SELECT * FROM Win32_BaseBoard")
    for obj_item in col_items:
        main_board = {
            'name': obj_item.Name,
            'description': obj_item.Description,
            'manufacturer': obj_item.Manufacturer,
            'model': obj_item.Model,
            'product': obj_item.Product,
            'serialNumber': obj_item.SerialNumber,
            'version': obj_item.Version
        }
        motherboard_details.append(main_board)
    return motherboard_details

#Function to get VGA devices
def get_gpu_details():
    gpu_details = []
    computer = wmi.WMI()
    gpu_info = computer.Win32_VideoController()
    for eachitem in gpu_info:
        each_gpu = {
            'name': eachitem.Name,
            'adapterRAM': eachitem.AdapterRAM,
            'description': eachitem.description,
            'pnpDeviceID': eachitem.PNPDeviceID,
            'systemName': eachitem.SystemName
        }
        gpu_details.append(each_gpu)
    return gpu_details

#Function to get monitor details
def get_monitor_details():
    monitor_details = []
    obj = wmi.WMI().Win32_PnPEntity(ConfigManagerErrorCode=0)
    displays = [x for x in obj if 'Monitor' in str(x)]
    for item in displays:
        if item.PNPClass == "Monitor":
            each_monitor = {
                'name': item.Caption,
                'manufacturer': item.Manufacturer,
                'hardwareID': item.HardwareID[0],
                'description': item.Description,
                'systemName':item.SystemName
            }
            monitor_details.append(each_monitor)
    return monitor_details

#Function to get cd_drive details
def get_cd_drive_details():
    cd_drive_details = []
    obj = wmi.WMI().Win32_PnPEntity(ConfigManagerErrorCode=0)
    cd_drive = [x for x in obj if 'CDROM' in str(x)]
    for item in cd_drive:
        each_cd_drive = {
            'name': item.Name,
            'description': item.Description,
            'hardwareID': item.HardwareID[0],
            'pnpDeviceID': item.pnpDeviceID,
            'manufacturer': item.Manufacturer,
            'systemName': item.systemName
        }
        cd_drive_details.append(each_cd_drive)
    return cd_drive_details

#Function to get mouse details
def get_mouse_details():
    mouse_details = []
    obj = wmi.WMI().Win32_PnPEntity(ConfigManagerErrorCode=0)
    mouse = [x for x in obj if 'Mouse' in str(x)]
    for item in mouse:
        each_mouse = {
            'name': item.Caption,
            'manufacturer': item.Manufacturer,
            'hardwareID': item.HardwareID[0],
            'description': item.Description,
            'systemName':item.SystemName
        }
        mouse_details.append(each_mouse)
    return mouse_details

#Fuction to get speaker details
def get_speaker_details():
    speaker_details = []
    obj = wmi.WMI().Win32_PnPEntity(ConfigManagerErrorCode=0)
    speakers = [x for x in obj if 'Speakers' in str(x)]
    for item in speakers:
        each_speaker = {
            'name': item.Caption,
            'manufacturer': item.Manufacturer,
            'hardwareID': item.HardwareID[0],
            'description': item.Description,
            'systemName':item.SystemName
        }
        speaker_details.append(each_speaker)
    return speaker_details

#Function to get keyboard details
def get_keyboard_details():
    keyboard_details = []
    obj = wmi.WMI().Win32_PnPEntity(ConfigManagerErrorCode=0)
    keyboard = [x for x in obj if 'Keyboard' in str(x)]
    for item in keyboard:
        each_keyboard = {
            'name': item.Caption,
            'manufacturer': item.Manufacturer,
            'hardwareID': item.HardwareID[0],
            'description': item.Description,
            'systemName':item.SystemName
        }
        keyboard_details.append(each_keyboard)
    return keyboard_details

#Function to get Hard Disk details
def get_hard_disk_details():
    hard_disk_details = []
    str_computer = "."
    obj_wmi_service = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    obj_swbem_services = obj_wmi_service.ConnectServer(str_computer, "root\cimv2")
    col_items = obj_swbem_services.ExecQuery("SELECT * FROM Win32_DiskDrive")
    for obj_item in col_items:
        each_hard_disk = {
            'name': obj_item.Name,
            'manufacturer': obj_item.Manufacturer,
            'deviceID': obj_item.DeviceID,
            'description': obj_item.Description,
            'pnpDeviceID': obj_item.PNPDeviceID,
            'size': obj_item.Size
        }
        hard_disk_details.append(each_hard_disk)
    return hard_disk_details

#Get RAM details
def get_ram_details():
    ram_details = []
    str_computer = "."
    obj_wmi_service = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    obj_swbem_services = obj_wmi_service.ConnectServer(str_computer, "root\cimv2")
    col_items = obj_swbem_services.ExecQuery("SELECT * FROM Win32_PhysicalMemory")
    for obj_item in col_items:
        each_ram = {
            'name': obj_item.Name,
            'manufacturer': obj_item.manufacturer,
            'description': obj_item.description,
            'serialNumber': obj_item.SerialNumber,
            'capaity': obj_item.capacity
        }
        ram_details.append(each_ram)
    return ram_details



import hashlib

REPLACEMENTS_CHARS = {
    'Ã‚Â°': '°',
    'Ã©': 'é',
    'Ãƒâ€°': 'É',
    'ÃƒÂ©': 'é',
    'ÃƒÂ¨': 'è',
    'ÃƒÂ ': 'à',
    'ÃƒÂ¢': 'â',
    'ÃƒÂ®': 'î',
    'ÃƒÂ´': 'ô',
    'ÃƒÂ»': 'û',
    'Ã¨': 'è',
    'Ã ': 'à',
    'Ã¢': 'â',
    'Ã®': 'î',
    'Ã´': 'ô',
    'Ã»': 'û'
}

def string2hash(input_string:str, algorithm:str='sha256') -> str:
    # Crée un nouvel objet de hash avec l'algorithme spécifié
    hash_function = hashlib.new(algorithm)
    # Met à jour l'objet de hash avec la chaîne de caractères (en encodage binaire)
    hash_function.update(input_string.encode('utf-8'))
    # Renvoie la chaîne de caractères hachée sous forme hexadécimale
    return hash_function.hexdigest()

def fix_encoding(text:str) -> str:
    """
    Corrige les problèmes d'encodage courants dans le texte.
    
    Args:
        text (str): Le texte mal encodé
        
    Returns:
        str: Le texte avec l'encodage corrigé
    """
    # Méthode 1: Utilisation de encode/decode
    try:
        # Essaie de décoder depuis latin1 puis encoder en utf-8
        return text.encode('latin1').decode('utf-8')
    except:
        pass
    
    # Méthode 2: Remplacement direct des séquences problématiques
    for bad, good in REPLACEMENTS_CHARS.items():
        text = text.replace(bad, good)
    
    return text

def multi_replace(texte:str, *replacements:list[tuple[str]]|list[str]) -> str:
    """
    Remplace plusieurs chaînes dans le texte.

    Args:
        texte: La chaîne de caractères à modifier.
        replacements: Un ou plusieurs tuples sous la forme (ancien, nouveau) ou (ancien, nouveau, occurrences) ou (ancien1, nouveau1, ancien2, nouveau2, ...).

    Returns:
        Une nouvelle chaîne avec les remplacements appliqués.
    """
    result = texte
    if isinstance(replacements[0], (list, tuple)) and len(replacements) == 1:
        for t in range(0, len(replacements[0]), 2):
            old, new = replacements[0][t], replacements[0][t+1]
            result = result.replace(old, new)
    elif isinstance(replacements[0], (list, tuple)):
        for t in replacements:
            old, new, occurrences = t[0], t[1], t[2] if len(t) > 2 else -1
            result = result.replace(old, new, occurrences)
    else:
        assert "type not found"
    return result

def multi_replace_by_one(texte:str, olds:tuple[str]|list[str], new:str) -> str:
    assert isinstance(olds, (list, tuple)), "'olds' is not a list or tuple of string"
    result = texte
    for old in olds:
        result = result.replace(old, new)
    return result

def while_replace(texte:str, old:str, new:str) -> str:
    result = texte
    while old in result:
        result = result.replace(old, new)
    return result



# __init__.py

"""
kelian: A Python library of useful code snippets
------------------------------------------------
Description:
    This library speeds up Python development by avoiding reinventing the wheel.
    It is made up of many commonly used code snippets, such as utility functions,
    classic algorithms, data manipulations and much more, to simplify development
    and improve productivity.

Auteur:
    Kelian

Licence:
    MIT License (voir LICENSE pour plus de détails)

Version:
    0.1.8
"""

__version__ = "0.1.8"
__author__ = "Kelian"
__license__ = "MIT"

from .encryption import alpha2dict, list2dict, encrypt, decrypt, encrypt_by_list, decrypt_by_list, encrypt_by_character_manga, decrypt_by_character_manga
from .loading_bar import ProgressBar
from .system import get_processor_details, get_motherboard_details, get_gpu_details, get_monitor_details, get_cd_drive_details, get_mouse_details, get_speaker_details, get_keyboard_details, get_hard_disk_details, get_ram_details
from .utils import string2hash, fix_encoding, multi_replace, while_replace

# Définir ce qui sera importé lors d'un 'from package import *'
__all__ = [
    "alpha2dict", 
    "list2dict", 
    "encrypt", 
    "decrypt", 
    "encrypt_by_list", 
    "decrypt_by_list", 
    "encrypt_by_character_manga", 
    "decrypt_by_character_manga", 
    "ProgressBar", 
    "get_processor_details", 
    "get_motherboard_details", 
    "get_gpu_details", 
    "get_monitor_details", 
    "get_cd_drive_details", 
    "get_mouse_details", 
    "get_speaker_details", 
    "get_keyboard_details", 
    "get_hard_disk_details", 
    "get_ram_details", 
    "string2hash",
    "fix_encoding", 
    "multi_replace", 
    "while_replace"
]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kelian",
    version="0.1.8",
    author="Kelian",
    description="Library of useful code snippets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Game-K-Hack/kelian",
    project_urls={
        "Bug Tracker": "https://github.com/Game-K-Hack/kelian/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3", 
        "License :: OSI Approved :: MIT License", 
        "Operating System :: OS Independent"
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.11",
    install_requires=[
        "pywin32", 
        "WMI"
    ]
)



import random
from .data.encryption.manga import CHARACTER_MANGA_LIST

ALPHA = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz0123456789éêï~\"#'{([-|è`_\çà@)]=}°/*+<>?,.;:§!¨^¤£$€µ%ù²& \n"
ALPHA_HEX = "0123456789abcdef"

def alpha2dict(x:str) -> dict:
    return {
        list(x)[i]:[i+1 for i in range(len(list(x)))][i]
        for i in range(len(list(x)))
    }

def list2dict(list_:list[str], alpha:str=ALPHA_HEX) -> tuple[dict, dict]:
    data_alpha = {k:[] for k in alpha}
    data_list_ = {k:[] for k in list_}

    if len(alpha) < len(list_):
        alpha *= (len(list_) // len(alpha)) + 1
        alpha = alpha[:len(list_)]
    else:
        list_ *= (len(alpha) // len(list_)) + 1
        list_ = list_[:len(alpha)]

    for index in range(len(alpha)):
        data_alpha[alpha[index]].append(list_[index])

    for index in range(len(list_)):
        data_list_[list_[index]].append(alpha[index])
    
    return data_alpha, data_list_

def encrypt(message:str, password:str, alpha:str=ALPHA) -> str:
    message = message.encode('utf-8').hex()
    dictionaire = alpha2dict(alpha)
    intmessage, intpassword, inttoken, strtoken = [], [], [], []
    while len(list(message)) > len(list(password)): password = password + password
    for i in range(len(password)): intpassword.append(dictionaire[password[i]])
    for i in range(len(message)): intmessage.append(dictionaire[message[i]])
    for i in range(len(message)): inttoken.append(intmessage[i] + intpassword[i])
    for i in range(len(inttoken)):
        if inttoken[i] > len(list(alpha)): strtoken.append(str(list(dictionaire.keys())[list(dictionaire.values()).index(inttoken[i]-len(list(alpha)))]))
        else: strtoken.append(str(list(dictionaire.keys())[list(dictionaire.values()).index(inttoken[i])]))
    return "".join(strtoken)

def decrypt(message:str, password:str, alpha:str=ALPHA) -> str:
    dictionaire = alpha2dict(alpha)
    intmessage, intpassword, inttoken, resultmessage = [], [], [], []
    while len(list(message)) > len(list(password)): password = password + password
    for i in range(len(password)): intpassword.append(dictionaire[password[i]])
    for i in range(len(message)): inttoken.append(dictionaire[message[i]])
    for i in range(len(message)): intmessage.append(inttoken[i] - intpassword[i])
    for i in range(len(intmessage)):
        if intmessage[i] <= 0: resultmessage.append(list(dictionaire.keys())[list(dictionaire.values()).index(intmessage[i]+len(alpha))])
        else: resultmessage.append(list(dictionaire.keys())[list(dictionaire.values()).index(intmessage[i])])
    return bytes.fromhex("".join(resultmessage)).decode()

def encrypt_by_list(message:str, password:str, list_:list[str], sep:str="") -> str:
    data, _ = list2dict(list_, ALPHA_HEX)
    password = password.encode('utf-8').hex()
    return sep.join([
        random.choice(data[char]) 
        for char in list(encrypt(message, password, ALPHA_HEX))
    ])

def decrypt_by_list(message:str, password:str, list_:list[str], sep:str="") -> str:
    _, data = list2dict(list_, ALPHA_HEX)
    password = password.encode('utf-8').hex()
    return decrypt("".join([
        data[char][0] 
        for char in message.split(sep)
    ]), password, ALPHA_HEX)

def encrypt_by_character_manga(message:str, password:str) -> str:
    return encrypt_by_list(message, password, CHARACTER_MANGA_LIST, "\n")

def decrypt_by_character_manga(message:str, password:str) -> str:
    return decrypt_by_list(message, password, CHARACTER_MANGA_LIST, "\n")



from types import NoneType
from average_time import AverageTime
from typing import Any

class ProgressBar():
    def __init__(self, total_or_data:int|Any, current:int=0, length:int=50, enabled_average_time:bool=True, hide_average_time_when_is_over:bool=True, enabled_display_in_iterator:bool=True) -> None:
        if isinstance(total_or_data, int):
            self.total:int = total_or_data
            self.data_iterator:Any = None
        else:
            self.total:int = len(total_or_data)+1
            self.data_iterator:Any = iter(total_or_data)
        self.current:int = current
        self.length:int = length
        self.pattern:tuple[str, str, str, str, bool, str, bool] = ("[", "=", ".", "]", True, " | ", True)
        self.enabled_average_time = enabled_average_time
        if enabled_average_time:
            self.average_time = AverageTime()
        self.hide_average_time_when_is_over:bool = hide_average_time_when_is_over
        self.enabled_display_in_iterator:bool = enabled_display_in_iterator

    def __iter__(self):
        return self

    def __next__(self):
        try:
            print(self.__str__(), end="\r")
            return next(self.data_iterator)
        except StopIteration:
            raise StopIteration

    def format(self, start_side:str=None, full:str=None, empty:str=None, end_side:str=None, percent:bool=None, separator:str=None, average_time:bool=None) -> None:
        assert isinstance(start_side, (str, NoneType)), "'start_side' is not of type string"
        assert isinstance(full, (str, NoneType)), "'full' is not of type string"
        assert isinstance(empty, (str, NoneType)), "'empty' is not of type string"
        assert isinstance(end_side, (str, NoneType)), "'end_side' is not of type string"
        assert isinstance(percent, (bool, NoneType)), "'percent' is not of type boolean"
        assert isinstance(separator, (str, NoneType)), "'separator' is not of type string"
        assert isinstance(average_time, (bool, NoneType)), "'average_time' is not of type boolean"

        self.pattern = (
            start_side if start_side is not None else self.pattern[0], 
            full if full is not None else self.pattern[1], 
            empty if empty is not None else self.pattern[2], 
            end_side if end_side is not None else self.pattern[3], 
            percent if percent is not None else self.pattern[4], 
            separator if separator is not None else self.pattern[5], 
            average_time if average_time is not None else self.pattern[6]
        )

    def update(self) -> None:
        self.current += 1
        if self.enabled_average_time:
            self.average_time.loop(self.average_time.DEFAULT_ID)

    def display(self, update:bool=False) -> str:
        if update:
            self.update()

        full_bar_value = (self.length*self.current) // self.total

        result = "\x1b[2K" # clear the current line
        result += self.pattern[0]
        result += self.pattern[1] * full_bar_value
        result += self.pattern[2] * (self.length - full_bar_value)
        result += self.pattern[3]

        if self.pattern[4]: # if percent
            percent_value = round((100*self.current)/self.total, 1)
            result += " " + " "*(5-len(str(percent_value))) + str(percent_value) + "%"

        if self.enabled_average_time and self.pattern[6]: # if average_time
            if self.current == self.total:
                if not self.hide_average_time_when_is_over:
                    result += self.pattern[5]
                    result += "0:00:00.0"
            else:
                result += self.pattern[5]
                avg = self.average_time.get_average(self.average_time.DEFAULT_ID)
                result += "" if str(avg).startswith("E: ") else str((self.total - self.current)*avg)

        return result

    def __str__(self) -> str:
        return self.display(True)

    def __repr__(self) -> str:
        return self.__str__()
